package org.processmining.plugins.DataConformance.DCnet;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Map.Entry;

import org.nfunk.jep.JEP;
import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;
import org.processmining.models.DCNet.DCNet;
import org.processmining.models.DCNet.DCNetBinding;
import org.processmining.models.DCNet.DCNetNode;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.plugins.DataConformance.Utility;
import org.processmining.plugins.DataConformance.framework.BelongsTo;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.framework.ExecutionTrace;

public class DCNetExecutionTrace extends ExecutionTrace 
{
	private DCNet model;
	private DCState state;
	private HashMap<String, Object> valDiffTauStep=null;	
	private DCNetExecutionTrace(int initialCapacity) { super(initialCapacity); }
	
	public DCNetExecutionTrace(DCNet model)
	{ 
		super(1);
		this.model=model;
		this.state=new DCState();
		this.state.addObligation(model.getStartNode(), DCNetNode.emptyNode);
	}

	private Set<HashMap<String,Object>> buildTupleValues(Set<String> variableSet)
	{
		String[] variables=variableSet.toArray(new String[0]);
		if (variables.length>0)
			return buildTupleValuesHelp(variables,0);
		else {
			HashSet<HashMap<String,Object>> aSet=new HashSet<HashMap<String,Object>>();
			aSet.add(new HashMap<String, Object>());
			return aSet;
		}
	}
	
	public boolean isComplete()
	{
		try 
		{
			return state.getPerformableActivities().size()==0;
		} catch (NoSuchElementException  e) 
		{
			return false;
		}
	}
	
	private DCNetExecutionTrace execute(CNetNode node, CNetBinding inputBinding, CNetBinding outputBinding, 
			Map<String, Object> writtenVariables)
	{
		DCNetExecutionTrace trace=(DCNetExecutionTrace) this.clone();
		trace.valDiffTauStep=null;
		trace.state.removeObligations(node, inputBinding.getBoundNodes());
		trace.state.addObligations(outputBinding.getBoundNodes(), node);
		trace.state.setDataVariables(writtenVariables);
		ExecutionStep step=new ExecutionStep(node.getLabel());
		step.putAll(writtenVariables);
		trace.add(step);
		return trace;		
	}
		
	@SuppressWarnings("unchecked")
	private Set<HashMap<String, Object>> buildTupleValuesHelp(String[] variables, int i) 
	{
		assert(i<variables.length);	
		if(i==(variables.length-1))
		{
			Set<Comparable> valuesForVariable=model.getSetValues(model.getSetForVariable(variables[i]));
			HashSet<HashMap<String,Object>> retValue=new HashSet<HashMap<String,Object>>();
			for (Object varValue : valuesForVariable)
			{
				HashMap<String,Object> newTuple=new HashMap<String, Object>();
				newTuple.put(variables[i], varValue);
				retValue.add(newTuple);				
			}
			return retValue;
		}
		else
		{
			Set<Comparable> valuesForVariable=model.getSetValues(model.getSetForVariable(variables[i]));
			Set<HashMap<String,Object>> partialTuples=buildTupleValuesHelp(variables, i+1); 
			HashSet<HashMap<String,Object>> retValue=new HashSet<HashMap<String,Object>>();
			for (Object varValue : valuesForVariable)
			{
				for(HashMap<String,Object> partialTuple : partialTuples)
				{
					HashMap<String,Object> newTuple=(HashMap<String, Object>) partialTuple.clone();
					newTuple.put(variables[i], varValue);
					retValue.add(newTuple);
				}
			}
			return(retValue);
		}
	}
	
	private DCNetNode getNode(ExecutionStep step)
	{
		if (step==null)
			return null;
		for(CNetNode activity : model.getNodes())
			if (activity.getLabel().equals(step.getActivity()))
				return (DCNetNode) activity;
		return null;
	}
	
	public Collection<ExecutionTrace> successors(ExecutionStep logStep)
	{
		LinkedList<ExecutionTrace> retValue=new LinkedList<ExecutionTrace>();
		HashSet<DCNetNode> nodes=new HashSet<DCNetNode>();
		DCNetNode activity=getNode(logStep);
		
		//If an activity is given as input, the process trace successors must include
		//all possible execution steps for the same activity. 
		//If no activity is not given as input (i.e. a TAU-step), then the execution step
		//may refer to any executable activity for any admissible writing of variables.
		if (activity!=null)
		{
			List<CNetNode> obligations=state.getObligationsPerActivity(activity);
			if (obligations==null)
				return retValue;
			else
				nodes.add(activity);
		}
		else
			nodes.addAll(state.getPerformableActivities());
		for(DCNetNode node : nodes) {
			Set<CNetBinding> inputs;
			inputs=model.getInputBindings(node);
			for(CNetBinding input : inputs)
			{
				DCNetBinding inputBinding=(DCNetBinding)input;
				if (state.getObligationsPerActivity(node).containsAll(inputBinding.getBoundNodes()))
				{
					JEP eval=new JEP();
					eval.setAllowUndeclared(true);
					eval.addFunction("belongsTo", BelongsTo.getInstance());	
					Node inputBindingCondition;
					try {
						inputBindingCondition = eval.parse(inputBinding.getDataExpression());
					} catch (ParseException e) {
						e.printStackTrace();
						continue;
					}
					for(CNetBinding output : model.getOutputBindings(node))
					{
						
						DCNetBinding outputBinding=(DCNetBinding)output;
						String enablingCondition="("+
						node.getGuard()+") && ("+outputBinding.getDataExpression()+")"; 
						
						Node dataAndOutputBindingCondition;
						try {
							dataAndOutputBindingCondition = eval.parse(enablingCondition);
						} catch (ParseException e) {
							e.printStackTrace();
							continue;
						}
						
						
						enablingCondition="("+inputBinding.getDataExpression()+") && ("+
						node.getGuard()+") && ("+outputBinding.getDataExpression()+")";
						
						Set<String> variableSet = DCNet.getWrittenVariables(enablingCondition);
						Set<HashMap<String,Object>> writtenVarSet = buildTupleValues(variableSet);		
						
						float minValDiff=Float.MAX_VALUE;
						if (model.areNonAffectingVariables(node))
							for(HashMap<String,Object> writtenVariable : writtenVarSet) {						
								if(evaluate(eval,inputBindingCondition,dataAndOutputBindingCondition,state.getDataVariables(),writtenVariable)) {
									DCNetExecutionTrace newTrace=execute(node, inputBinding, outputBinding, writtenVariable);
									retValue.add(newTrace);
								}
							}
						else
						{
							valDiffTauStep=null;
							HashSet<ExecutionTrace> minValDiffSet=new HashSet<ExecutionTrace>();
							for(HashMap<String,Object> writtenVariable : writtenVarSet)
							{
								boolean evaluation=evaluate(eval,inputBindingCondition,dataAndOutputBindingCondition,state.getDataVariables(),writtenVariable);
								if(evaluation)
								{
									float valDiff=getValDiff(writtenVariable,logStep,node);
									if(valDiff<=minValDiff)
									{
										DCNetExecutionTrace newTrace=execute(node, inputBinding, outputBinding, writtenVariable);
										if (valDiff<minValDiff)
											minValDiffSet.clear();
										minValDiffSet.add(newTrace);
										minValDiff=valDiff;
									}
								}
							}
							retValue.addAll(minValDiffSet);
						}
					}
				}
			}
		}
		return retValue;
	}
	
	

	private float getValDiff(HashMap<String, Object> writtenVariable, ExecutionStep step, DCNetNode node) 
	{
		if (valDiffTauStep==null) 
			valDiffTauStep=new HashMap<String, Object>();
		Set<String> varName=model.getNonAffectingVariables(node);
		if (varName==null)
			return Float.MAX_VALUE;
		else if (step.equals(ExecutionStep.bottomStep)) {
			if (valDiffTauStep.size()==0) {
				for(String var : varName) 
					valDiffTauStep.put(var, writtenVariable.get(var));
				return 0;
			}
			else {
				for(String var : varName)
					if (!valDiffTauStep.get(var).equals(writtenVariable.get(var)))
						return Float.MAX_VALUE;
				return 0;
			}		
		}
		else
		{
			float retValue=-1;
			for(Entry<String, Object> entryVar : writtenVariable.entrySet())
			{
				if (varName.contains(entryVar.getKey()) && !step.containsKey(entryVar.getKey())) {
					if (valDiffTauStep.get(entryVar.getKey()) == null) 
						valDiffTauStep.put(entryVar.getKey(), entryVar.getValue());
					else  
						if (!valDiffTauStep.get(entryVar.getKey()).equals(entryVar.getValue())) {
							return Float.MAX_VALUE;
						}

				}
				if (varName.contains(entryVar.getKey()) && step.containsKey(entryVar.getKey()))
				{
					if (retValue==-1) retValue=0;
					if (entryVar.getValue() instanceof Number && step.get(entryVar.getKey()) instanceof Number)
						retValue+=Math.abs(((Number)entryVar.getValue()).floatValue()-((Number)step.get(entryVar.getKey())).floatValue());
					else if (entryVar.getValue() instanceof String && step.get(entryVar.getKey()) instanceof String)
						retValue+=Utility.stringDistance((String)entryVar.getValue(), (String)step.get(entryVar.getKey()));
					else retValue+=(entryVar.equals(step.get(entryVar.getKey()))) ? 0 : 1 ;
				}
			}
			if (retValue==-1)
				return Float.MAX_VALUE;
			else
				return retValue;
		}
	}

	private boolean evaluate(JEP eval, Node inputBindingCondition, Node dataAndOutputBindingCondition,
			Map<String, Object> readVariables, HashMap<String, Object> writtenVariables) 
	{
		eval.getSymbolTable().clearValues();
		for(Entry<String, Object> entry : readVariables.entrySet()) {
			eval.addVariable(entry.getKey(), entry.getValue());
		}
		for(Entry<String, Object> entry : writtenVariables.entrySet()) 
		{
			String varName=entry.getKey();
			if(DCNet.OUTPUTVAR_IDENTIFIER_POS==DCNet.Position.AFTER)
				varName+=DCNet.OUTPUTVAR_IDENTIFIER;
			else
				varName=DCNet.OUTPUTVAR_IDENTIFIER+varName;
			eval.addVariable(varName, entry.getValue());
		}
		try
		{
			if (eval.evaluate(inputBindingCondition).equals(new Double(0)))
				return false;
			else
				return (eval.evaluate(dataAndOutputBindingCondition).equals(new Double(1)));
		}
		catch(ParseException err)
		{
			err.printStackTrace();
			return false;
		}
	}

	
	public Object clone()
	{
		DCNetExecutionTrace clone=new DCNetExecutionTrace(this.size()+1);
		clone.addAll(this,numberOfBottomSteps);
		clone.model=model;
		clone.state=(DCState) state.clone();
		return clone;
	}

	public int hashCode() {
		return state.hashCode();
	}

	public boolean equals(Object obj) 
	{
		if (obj instanceof DCNetExecutionTrace) {
			DCNetExecutionTrace other=(DCNetExecutionTrace) obj;
			boolean sameState=this.state.equals(other.state);	
			return this.size()==other.size() && this.state.equals(other.state);
		}
		else
			return false;
	}

	public String getTraceName() {
		return null;
	}
	
}
