package org.processmining.plugins.DataConformance.DataAlignment.DCNet;

import java.text.ParseException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.models.DCNet.DCNet;
import org.processmining.models.DCNet.DCNetNode;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.connections.cnet.EvClassLogCNetConnection;
import org.processmining.plugins.DataConformance.DCnet.ResultReplayDCNet;
import org.processmining.plugins.DataConformance.GUI.MappingPanel;
import org.processmining.plugins.DataConformance.GUI.VariableMatchCostPanel;
import org.processmining.plugins.DataConformance.framework.ReplayState;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.cnet.replayresult.CNetRepResult;
import org.processmining.plugins.connectionfactories.logcnet.CNetEvClassMapping;

public class DataAlignment {
	@Plugin(name = "Conformance Checking of Data-aware Casual Net (using ILP)",
			parameterLabels = { "Data-aware Casual Net", "Replay Result", "Log"},
			returnLabels = { "Result Replay"},
			returnTypes = { ResultReplayDCNet.class },
			userAccessible = true,
			categories = { PluginCategory.ConformanceChecking },
			help = "Data-flow Alignment")
			@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "M. de Leoni", email = "m.d.leoni@tue.nl", pack = "DataAwareReplayer")
	public static ResultReplayDCNet dataConformance(UIPluginContext context, DCNet net, CNetRepResult input, XLog log) 
		throws ParseException, InterruptedException 
	{
		XLogInfo summary = XLogInfoFactory.createLogInfo(log);
		XEventClasses eventClasses=summary.getEventClasses();
		
		//Activity Mapping
		EvClassLogCNetConnection conn = null;

		try {
			conn = context.getConnectionManager().getFirstConnection(EvClassLogCNetConnection.class, context, net,
					log);
		} catch (Exception e) {
			e.printStackTrace();
			context.log(e.getMessage());
		}


		Set<Entry<CNetNode, XEventClass>> entrySet = ((CNetEvClassMapping) conn.getObjectWithRole(EvClassLogCNetConnection.CNET2EVCLASSMAPPING)).entrySet();		
		HashMap<DCNetNode,XEventClass> activityMapping=new HashMap<DCNetNode, XEventClass>();
		for(Entry<CNetNode, XEventClass> entry : entrySet)
		{
			activityMapping.put((DCNetNode) entry.getKey(),entry.getValue());
		}

		//Variable Mapping
		Collection<String> logAttributes=summary.getEventAttributeInfo().getAttributeKeys();
		TreeSet<String> processAttributes=new TreeSet<String>(net.getVariables());

		MappingPanel<String, String> mapVariablePanel=
			new MappingPanel<String, String>(processAttributes,logAttributes);
		InteractionResult result = context.showConfiguration("Setup Variable Mapping", mapVariablePanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		Map<String, String> variableMapping=mapVariablePanel.getMapping(false);	


		//Variable Costs
		VariableMatchCostPanel<XEventClass> variablePanel=new VariableMatchCostPanel<XEventClass>(activityMapping,mapVariablePanel.getMapping(true));
		result = context.showConfiguration("Variable Mapping Cost", variablePanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		VariableMatchCosts variableCost=variablePanel.getCosts();	

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		Collection<ReplayState> list=MultiThreadedDataAlignment.perform(net, log, input, activityMapping, eventClasses, variableMapping,variableCost,context);

		return new ResultReplayDataAlignment(list, variableCost, net, log,variableMapping, eventClasses.getClassifier());

	}


}

