package org.processmining.plugins.DataConformance.DataAlignment;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;

public class DataAlignerBuilder implements Runnable
{

	public static Collection<DataAlignmentState> buildAlignments(Collection<ControlFlowAlignment> listOfLists, VariableMatchCosts variableCost, Map<String,Class> varType, 
			Map<String,Object> upperBounds, Map<String,Object> lowerBounds, DataAlignerListener listener) {
		int maxConcurrentThreads=Runtime.getRuntime().availableProcessors();
		ThreadPoolExecutor pool = new ThreadPoolExecutor(maxConcurrentThreads, maxConcurrentThreads, 60, TimeUnit.SECONDS,new LinkedBlockingQueue<Runnable>());
		pool.allowCoreThreadTimeOut(true);
		Collection<DataAlignmentState> retValue=new ArrayList<DataAlignmentState>(listOfLists.size());
		for(IControlFlowAlignment<Object, Object> list : listOfLists)
		{
			//pool.execute(new DataAlignerBuilder(list, variableCost, varType, upperBounds, lowerBounds, listener,retValue));	
			new DataAlignerBuilder(list, variableCost, varType, upperBounds, lowerBounds, listener,retValue).run();
		}
		pool.shutdown();	
		try {
			pool.awaitTermination(10, TimeUnit.DAYS);
		} catch (InterruptedException e) {
			listener.exceptionThrow(e);
		}
		return retValue;
	}

	private Collection<DataAlignmentState> retValue;
	private DataAlignerListener listener;
	private Map<String, Object> lowerBounds;
	private Map<String, Object> upperBounds;
	private Map<String, Class> varType;
	private VariableMatchCosts variableCost;
	private IControlFlowAlignment<Object, Object> list;
	
	private DataAlignerBuilder(IControlFlowAlignment<Object, Object> list, VariableMatchCosts variableCost, 
			Map<String,Class> varType, Map<String,Object> upperBounds, Map<String,Object> lowerBounds, DataAlignerListener listener, Collection<DataAlignmentState> retValue)
	{
		this.list=list;
		this.variableCost=variableCost;
		this.varType=varType;
		this.upperBounds=upperBounds;
		this.lowerBounds=lowerBounds;
		this.listener=listener;
		this.retValue=retValue;	
	}
	
	public void run() {
		try {
			listener.startNewAlignment(list.getTraceName());
			DataAlignmentState alignment=
				SingleTraceDataAlignmentBuilder.createAlignment(list, variableCost, varType, upperBounds, lowerBounds);
			alignment.setControlFlowFitness(list.getControlFlowFitness());
			synchronized(retValue)
			{
				retValue.add(alignment);
			}
			listener.newAlignedTrace(alignment);
		} catch (Exception e) {
			e.printStackTrace();
			listener.exceptionThrow(list.getTraceName(),e);
		}
		
	}

}
