package org.processmining.plugins.DataConformance.DataAlignment;

import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.framework.ReplayState;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;

/**
 * Static utility methods for data alignment that are common among a few classes
 * 
 * @author fmannhardt
 *
 */
public class DataAlignmentHelper {
	
	public static float computeFitness(ReplayState state, VariableMatchCosts variableCost)
	{
		float numerator=state.getCost();
		float denumerator=0;
		for(ExecutionStep step : state.getProcessTracePrefix())
		{
			for(String variable : step.keySet())
				denumerator+=Math.max(variableCost.costFaultyValue(step.getActivity(), variable), 
						variableCost.notWritingCost(step.getActivity(), variable));	
		}
		float dataFlowFitness;
		if (denumerator>0)
			dataFlowFitness=(1-numerator/denumerator);
		else
			dataFlowFitness=1;
		if (state.getCost() == Float.MAX_VALUE) {
			return 0;
		}
		float controlFlowFitness=((DataAlignmentState)state).getControlFlowFitness();
		return (dataFlowFitness+controlFlowFitness)/2;
		
	}
	
	public static float computeMaxDataCost(ReplayState state, VariableMatchCosts variableCost) {
		float denumerator=0;
		for(ExecutionStep step : state.getProcessTracePrefix())
		{
			for(String variable : step.keySet())
				denumerator+=Math.max(variableCost.costFaultyValue(step.getActivity(), variable), 
						variableCost.notWritingCost(step.getActivity(), variable));	
		}
		return denumerator;
	}

}
