package org.processmining.plugins.DataConformance.RepairLog;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension.StandardModel;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeBooleanImpl;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.deckfour.xes.model.impl.XAttributeDiscreteImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeTimestampImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.ProMCheckComboBox;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.Alignment.AlignmentStep;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.petrinetwithdata.decisionminer.MyPropertyPanel;

public class RepairLog {
	
	
	
	@Plugin(name = "Repair Log With Respect to Alignment",
			level = PluginLevel.PeerReviewed,
			parameterLabels = { "Alignments", "Petri net with data"},
			returnLabels = { "Repaired Event Log"},
			returnTypes = { XLog.class },
			userAccessible = true,
			categories = { PluginCategory.Enhancement },
			help = "Given a model M and a data-aware alignment A created starting from a log L, "
					+ "it will fix log L by generating a log L2 so that the alignment of L2 and M will"
					+ " not contain moves in log, model or synchronous moves without "
					+ "incorrect write operations of selected types."
			
			)
			@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "M. de Leoni", email = "m.d.leoni@tue.nl", pack = "DataAwareReplayer")
			public static XLog plugin(UIPluginContext context, ResultReplay alignments,DataPetriNet net)
	{
		XFactory factory =  XFactoryRegistry.instance().currentDefault();
		XLog initialLog = alignments.getAlignedLog();
		XLog repairedLog=factory.createLog();
		context.getProgress().setMaximum(alignments.labelStepArray.size());
		HashMap<String, XTrace> map=new HashMap<String, XTrace>();
		for(XTrace t : initialLog)
			map.put(XConceptExtension.instance().extractName(t), t);
		
		SortedSet<String> logMoves=new TreeSet<String>();
		SortedSet<String> modelMoves=new TreeSet<String>();
		SortedSet<String> synchNoGoodMoves=new TreeSet<String>();

		
		for (Alignment alignment : alignments.labelStepArray)
		{	
			Iterator<AlignmentStep> iterator = alignment.iterator();
			while(iterator.hasNext())
			{
				AlignmentStep align=iterator.next();
				switch(align.getType())
				{
					case L :
						logMoves.add(align.getLogView().getActivity());
						break;
					case LMNOGOOD :
						synchNoGoodMoves.add(align.getProcessView().getActivity());
						break;
					case MREAL :
						modelMoves.add(align.getProcessView().getActivity());
						break;
					default :
						break;					
				}
			}
		}
		
		MyPropertyPanel panel=new MyPropertyPanel("");
		ProMCheckComboBox logMovesCombo=new ProMCheckComboBox(logMoves);
		logMovesCombo.addSelectedItems(logMoves);
		ProMCheckComboBox modelMovesCombo=new ProMCheckComboBox(modelMoves);
		modelMovesCombo.addSelectedItems(modelMoves);
		ProMCheckComboBox synchrNoGoodMovesCombo=new ProMCheckComboBox(synchNoGoodMoves);
		synchrNoGoodMovesCombo.addSelectedItems(synchNoGoodMoves);
		
		panel.addProperty("Log moves:", logMovesCombo);
		panel.addProperty("Process moves:", modelMovesCombo);
		panel.addProperty("Synchronous moves with wrong write operations:", synchrNoGoodMovesCombo);
		
		InteractionResult result = context.showConfiguration("Moves to consider when repairing logs", panel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null; 
		}
	
		for (Alignment alignment : alignments.labelStepArray)
		{	
			XTrace newTrace=factory.createTrace();
			XTrace existingTrace=map.get(alignment.getTraceName());
			XConceptExtension.instance().assignName(newTrace, alignment.getTraceName());
			Iterator<XEvent> existingTraceIter=existingTrace.iterator();
			Iterator<AlignmentStep> alignIter = alignment.alignmentStepIterator();
			XEvent eventToAdd = null;
			
			while(alignIter.hasNext())
			{
				AlignmentStep step = alignIter.next();
				
				switch(step.getType())
				{
					case L :
						eventToAdd=createEvent(null,step);
						existingTraceIter.next();
						try {
							if (logMovesCombo.getSelectedItems()==null || logMovesCombo.getSelectedItems().contains(step.getLogView().getActivity()))
								eventToAdd=null;
						} catch (Exception e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						break;
					case LMGOOD :
						eventToAdd=createEvent(existingTraceIter.next(),step);
						break;
					case LMNOGOOD :
						XEvent nextEvent = existingTraceIter.next();
						eventToAdd=createEvent(nextEvent,step);
						if (synchrNoGoodMovesCombo.getSelectedItems()==null || synchrNoGoodMovesCombo.getSelectedItems().contains(step.getProcessView().getActivity()))
						{
							eventToAdd=createEvent(nextEvent,step);
						}
						break;
					case MREAL :
						if (modelMovesCombo.getSelectedItems()==null || modelMovesCombo.getSelectedItems().contains(step.getProcessView().getActivity()))
							eventToAdd=createEvent(null,step);
						else
							eventToAdd=null;
						break;
					case MINVI :
							eventToAdd=null;
							break;
					default:
						assert(false);
						break;
				}
				if (eventToAdd!=null)
					newTrace.add(eventToAdd);
			}
			if (newTrace.size()>0)
				repairedLog.add(addMissingTimeStamp(newTrace));
			//repairedLog.add(newTrace);
			context.getProgress().inc();
		}
		return repairedLog;
	}
	
	private static XTrace addMissingTimeStamp(XTrace aTrace) {
		int i=-1;
		XTrace newTrace=XFactoryRegistry.instance().currentDefault().createTrace(aTrace.getAttributes());
		Date datePreviousEvent=null;
		do {
			datePreviousEvent=XTimeExtension.instance().extractTimestamp(aTrace.get(++i));
		} while(datePreviousEvent==null && i<aTrace.size());
		if (datePreviousEvent==null)
		{
			datePreviousEvent=new Date();
			System.out.println("No timestamp for trace "+XConceptExtension.instance().extractName(aTrace));
		}
		for(int j=0;j<i;j++)
		{
			XEvent anEvent=aTrace.get(j);
			XTimeExtension.instance().assignTimestamp(anEvent, (long) (datePreviousEvent.getTime()/Math.pow(1.001, i-j)));
			newTrace.add(anEvent);
		}
		for(;i<aTrace.size();i++)
		{
			Date newDateEvent=XTimeExtension.instance().extractTimestamp(aTrace.get(i));
			if (newDateEvent==null)
			{
				for(int j=i+1;newDateEvent==null & j<aTrace.size();j++)
					newDateEvent=XTimeExtension.instance().extractTimestamp(aTrace.get(j));	
				if(newDateEvent==null)
					newDateEvent=new Date((long) (datePreviousEvent.getTime()*1.001));
				else
					newDateEvent=new Date((datePreviousEvent.getTime()+newDateEvent.getTime())/2);
				datePreviousEvent=newDateEvent;
				XEvent anEvent=aTrace.get(i);
				XTimeExtension.instance().assignTimestamp(anEvent, newDateEvent);
				newTrace.add(anEvent);
			}
			else
				newTrace.add(aTrace.get(i));
		}
		return newTrace;
	}
	


	private static XEvent createEvent(XEvent event,AlignmentStep step) {
		String activityName=step.getProcessView().getActivity();
		if (activityName==null)
			activityName=step.getLogView().getActivity();
		XEvent newEvent=XFactoryRegistry.instance().currentDefault().createEvent();
		if (event==null)
		{
			XLifecycleExtension.StandardModel activityEventType = StandardModel.COMPLETE;
			XLifecycleExtension.instance().assignStandardTransition(newEvent, activityEventType);
		}
		else
		{
			
			XLifecycleExtension.instance().assignStandardTransition(newEvent, XLifecycleExtension.instance().extractStandardTransition(event));
			XTimeExtension.instance().assignTimestamp(newEvent, XTimeExtension.instance().extractTimestamp(event));
		}
		XConceptExtension.instance().assignName(newEvent, activityName);
		
		for(Entry<String, Object> entry : step.getProcessView().entrySet())
		{
			String key=entry.getKey();
			Object value=entry.getValue();
			XAttribute attrib;
			if (value instanceof Boolean)
				attrib=new XAttributeBooleanImpl(key,(Boolean) value);
			else if (value instanceof Integer)
				attrib=new XAttributeDiscreteImpl(key, (Integer) value);
			else if (value instanceof Float || value instanceof Double)
				attrib=new XAttributeContinuousImpl(key, (Double) value);
			else if (value instanceof Date)
				attrib=new XAttributeTimestampImpl(key, (Date) value);
			else
				attrib=new XAttributeLiteralImpl(key, value.toString());
			newEvent.getAttributes().put(key, attrib);
		}
		return newEvent;
	}

}
