package org.processmining.plugins.DataConformance.visualization;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.regex.Pattern;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.ReplayResultVisController.AlignmentFilter;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentLegend;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentCollectionVisualizer;

import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * Old visualization using the "wedges" from the PNetReplayer, is now being replaced by {@link AlignmentCollectionVisualizer}.
 *
 */
//@Plugin(name = "2 Model-Log Alignment in Presence of Data", returnLabels = { "Model-Log Alignment in Presence of Data" }, returnTypes = { JComponent.class }, parameterLabels = { "Matching Instances" }, userAccessible = false)
//@Visualizer
@Deprecated
public class ReplayResultVis {

	private static final int LIMIT = 1000;
	private static final String ALIGNMENT_PANEL_TITLE = "Log-Model alignments";
	private static final String SHOW_NEXT_TEXT = "Show next %d alignments";
	
	private ProMSplitPane northSplitPanel;

	@PluginVariant(requiredParameterLabels = { 0 })
	public JComponent visualize(PluginContext context, ResultReplay logReplayResult) {
		return doVisualize(logReplayResult.labelStepArray);
	}

	public JComponent doVisualize(Iterable<Alignment> alignments) {
		JPanel alignmentPanel = buildAlignmentsPanel(alignments.iterator(), new AlignmentFilter() {
			
			public boolean showAlignment(Alignment align) {
				return true;
			}
		}, LIMIT);
		northSplitPanel = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		northSplitPanel.setLeftComponent(alignmentPanel);
		northSplitPanel.setResizeWeight(1.0);
		northSplitPanel.setOneTouchExpandable(true);
		
		JPanel rightPanel = new JPanel();
		rightPanel.setLayout(new BoxLayout(rightPanel, BoxLayout.Y_AXIS));
		rightPanel.add(new AlignmentLegend());
		rightPanel.add(createButtonPanel(alignments));
		northSplitPanel.setRightComponent(rightPanel);
		
		return northSplitPanel;
	}

	private JPanel buildAlignmentsPanel(Iterator<Alignment> alignmentIter, AlignmentFilter filter, int limit) {
		final JPanel alignmentPanel = SlickerFactory.instance().createRoundedPanel();
		alignmentPanel.setLayout(new GridBagLayout());
		final ProMPropertiesPanelWithComp alignmentPropertyPanel = new ProMPropertiesPanelWithComp(ALIGNMENT_PANEL_TITLE);
		
		// Show initial alignments
		final ReplayResultVisController apController = new ReplayResultVisController(alignmentPropertyPanel, alignmentIter, limit, filter);		
		apController.showNextAlignments();		
		alignmentPanel.add(alignmentPropertyPanel, getAlignmentBagConstraints());

		if (apController.hasMore()) {
			final JButton showMoreBtn = SlickerFactory.instance().createButton(String.format(SHOW_NEXT_TEXT, apController.getLimit()));
			showMoreBtn.addActionListener(new ActionListener() {
	
				public void actionPerformed(ActionEvent e) {
					apController.showNextAlignments();
					alignmentPropertyPanel.getParent().validate();
					if (apController.hasMore()) {
						alignmentPanel.getParent().validate();
					}
				}
			});
			alignmentPanel.add(showMoreBtn, getButtonBagConstraints());
		}
		return alignmentPanel;
	}

	private static GridBagConstraints getButtonBagConstraints() {
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.CENTER;
		c.gridx = 0;
		c.gridy = 1;
		return c;
	}

	private static GridBagConstraints getAlignmentBagConstraints() {
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		return c;
	}
	
	private JPanel createButtonPanel(final Iterable<Alignment> alignments) {
		final JPanel buttonPanel = SlickerFactory.instance().createRoundedPanel();
		buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.Y_AXIS));
		buttonPanel.add(SlickerFactory.instance().createLabel("Filter"));

		addFilterButton(alignments, buttonPanel, "Reset", new AlignmentFilter() {

			public boolean showAlignment(Alignment align) {
				return true;
			}
		});

		final JPanel gotoPaneByName = SlickerFactory.instance().createRoundedPanel();
		gotoPaneByName.setLayout(new GridLayout(1, 2));
		JButton gotoTraceBtn = SlickerFactory.instance().createButton("by trace name");
		final ProMTextField gotoTraceTextBox = new ProMTextField();
		gotoTraceBtn.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				final String traceRegex = gotoTraceTextBox.getText().trim();
				final Pattern regExPattern = Pattern.compile(traceRegex);
				northSplitPanel.setLeftComponent(buildAlignmentsPanel(alignments.iterator(), new AlignmentFilter() {

					public boolean showAlignment(Alignment align) {
						return regExPattern.matcher(align.getTraceName()).find();
					}

				}, 1000));
				northSplitPanel.getParent().validate();				
			}

		});
		gotoPaneByName.add(gotoTraceBtn);
		gotoPaneByName.add(gotoTraceTextBox);
		buttonPanel.add(gotoPaneByName);

		final JPanel gotoPaneByEventName = SlickerFactory.instance().createRoundedPanel();
		gotoPaneByEventName.setLayout(new GridLayout(1, 2));
		JButton gotoEventNameBtn = SlickerFactory.instance().createButton("by event name");
		final ProMTextField gotoEventNameTxt = new ProMTextField();
		gotoEventNameBtn.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				final String traceRegex = gotoEventNameTxt.getText().trim();
				final Pattern regExPattern = Pattern.compile(traceRegex);
				northSplitPanel.setLeftComponent(buildAlignmentsPanel(alignments.iterator(), new AlignmentFilter() {

					public boolean showAlignment(Alignment align) {
						for (ExecutionStep step : align.getProcessTrace()) {
							if (step.getActivity() != null) {
								if (regExPattern.matcher(step.getActivity()).find()) {
									return true;
								}
							}
						}
						for (ExecutionStep step : align.getLogTrace()) {
							if (step.getActivity() != null) {
								if (regExPattern.matcher(step.getActivity()).find()) {
									return true;
								}
							}
						}
						return false;
					}

				}, 1000));
				northSplitPanel.getParent().validate();				
			}

		});
		gotoPaneByEventName.add(gotoEventNameBtn);
		gotoPaneByEventName.add(gotoEventNameTxt);
		buttonPanel.add(gotoPaneByEventName);

		return buttonPanel;
	}
	
	private void addFilterButton(final Iterable<Alignment> alignments, final JPanel buttonPanel, String btnName,
			final AlignmentFilter filter) {
		JButton resetBtn = SlickerFactory.instance().createButton(btnName);
		resetBtn.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				northSplitPanel.setLeftComponent(buildAlignmentsPanel(alignments.iterator(), filter, LIMIT));
				northSplitPanel.getParent().validate();
			}

		});
		buttonPanel.add(resetBtn);
	}

}
