package org.processmining.plugins.DataConformance.visualization.alignment;

import java.awt.Color;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Map;

import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;

import com.google.common.collect.ComparisonChain;
import com.google.common.collect.Ordering;

public class AlignmentListView extends ProMTraceList<XAlignment> {

	public enum XAlignmentOrdering implements Comparator<XAlignment> {
		FITNESS_ASC("Fitness (Ascending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				return ComparisonChain.start().compare(a.getFitness(), b.getFitness()).result();
			}
		},
		FITNESS_DESC("Fitness (Descending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				return Collections.reverseOrder(FITNESS_ASC).compare(a, b);
			}
		},
		LENGTH_ASC("Length (Ascending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				return ComparisonChain.start().compare(a.size(), b.size()).result();
			}
		},
		LENGTH_DESC("Length (Descending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				return Collections.reverseOrder(LENGTH_ASC).compare(a, b);
			}
		},
		NAME_ASC("Trace Name (Ascending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				String nameA = a.getName();
				String nameB = b.getName();
				return ComparisonChain.start().compare(nameA != null ? nameA : "", nameB != null ? nameB : "").result();
			}
		},
		NAME_DESC("Trace Name (Descending)") {
			@Override
			public int compare(final XAlignment a, final XAlignment b) {
				return Collections.reverseOrder(NAME_ASC).compare(a, b);
			}
		};
		;

		private String description;

		private XAlignmentOrdering(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	}

	private static final long serialVersionUID = 9384229294683749L;

	public AlignmentListView(Collection<XAlignment> traces, Map<String, Color> activityColorMap) {
		this(traces, Ordering.from(XAlignmentOrdering.FITNESS_DESC), activityColorMap);
	}

	public AlignmentListView(Collection<XAlignment> traces, Comparator<XAlignment> order,
			final Map<String, Color> activityColorMap) {
		super(traces, new TraceBuilder<XAlignment>() {

			public Trace<Event> build(XAlignment a) {
				return new AlignmentTrace(a, activityColorMap);
			}
		}, order);
	}

	public AlignmentListView(Collection<XAlignment> traces, Comparator<XAlignment> order,
			TraceBuilder<XAlignment> traceBuilder) {
		super(traces, traceBuilder, order);
	}

}
