package org.processmining.plugins.DataConformance.visualization.alignment;

import javax.swing.JComponent;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentView.Layout;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Function;
import com.google.common.collect.Iterables;

@Plugin(name = "Log with AlignmentExtension", level = PluginLevel.Regular, //
 	returnLabels = { "Model-Log Alignment in Presence of Data" }, //
 	returnTypes = { JComponent.class }, parameterLabels = { "Matching Instances" })
@Visualizer
public class XAlignmentExtensionVisualizer {

	private AlignmentView dataAlignmentView;

	@PluginVariant(variantLabel = "Log with AlignmentExtension", requiredParameterLabels = { 0 })
	public JComponent visualize(PluginContext context, XLog extendedLog) {
		if (hasAlignmentExtension(extendedLog)) {
			return doVisualize(context, new XTraceResolver() {

				public XTrace getOriginalTrace(String name) {
					return null;
				}

				public boolean hasOriginalTraces() {
					return false;
				}

			}, Iterables.transform(extendedLog, new Function<XTrace, XAlignment>() {

				public XAlignment apply(XTrace trace) {
					return XAlignmentExtension.instance().extendTrace(trace);
				}
			}));
		} else {
			return SlickerFactory.instance().createLabel("XAlignmentExtension is missing!");
		}
	}

	private static boolean hasAlignmentExtension(XLog extendedLog) {
		boolean hasExtensionOnLogLevel = extendedLog.getExtensions().contains(XAlignmentExtension.instance());
		// As the extension is not correctly parsed when the XAlignmentExtension class was not yet loaded 
		boolean hasAlignmentAttribute = XAlignmentExtension.instance().extractFitness(extendedLog) != null;
		return hasExtensionOnLogLevel || hasAlignmentAttribute;
	}

	public JComponent doVisualize(PluginContext context, final XTraceResolver traceResolver,
			Iterable<XAlignment> alignments) {
		dataAlignmentView = new AlignmentView(Layout.TWOCOLUMN, context, traceResolver, ColorTheme.createColorMap(alignments));
		dataAlignmentView.getListView().addAll(alignments);
		return dataAlignmentView;
	}

}
