package org.processmining.plugins.DataConformance.visualization.grouping;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.concurrent.ExecutionException;

import javax.swing.JOptionPane;
import javax.swing.SwingWorker;

import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.traceview.masterdetail.MasterView;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.DataConformance.visualization.grouping.GroupedAlignments.AlignmentGroup;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;

public final class ActionExportAsAlignment implements ActionListener {

	private final PluginContext context;
	private final MasterView<AlignmentGroup, XAlignment> masterView;

	public ActionExportAsAlignment(PluginContext context, MasterView<AlignmentGroup, XAlignment> masterView) {
		this.context = context;
		this.masterView = masterView;
	}

	public void actionPerformed(ActionEvent e) {
		@SuppressWarnings("deprecation")
		final Object[] selection = masterView.getMasterList().getList().getSelectedValues();

		String proposedName = selection.length == 1 ? ((AlignmentGroup) selection[0]).getName() : "Group";
		final String groupName = JOptionPane.showInputDialog(masterView.getMasterList(),
				"Specify the name of the new ProM alignment object!", proposedName);

		if (groupName != null) {

			masterView.getMasterList().beforeUpdate();

			new SwingWorker<XLog, Void>() {

				protected XLog doInBackground() throws Exception {
					XLog groupLog = XFactoryRegistry.instance().currentDefault().createLog();

					groupLog.getExtensions().add(XAlignmentExtension.instance());
					double sumFitness = 0.0;
					int numAlignments = 0;

					int groupIndex = 0;
					for (Object o : selection) {
						AlignmentGroup group = (AlignmentGroup) o;
						Collection<XAlignment> alignments = masterView.getDetailElements(group);
						for (XAlignment a : alignments) {
							XTrace clonedTrace = (XTrace) a.getTrace().clone();
							addMetaData(clonedTrace, groupIndex, group.size());
							groupLog.add(clonedTrace);
							sumFitness += a.getFitness();
							numAlignments++;
						}
						groupIndex++;
					}
					XAlignmentExtension.instance().assignFitness(groupLog, sumFitness / numAlignments);

					return groupLog;
				}

				private void addMetaData(XTrace detailTraces, int groupIndex, int groupCount) {
					XUtils.putAttribute(detailTraces,
							XUtils.createAttribute("variant:index", Long.valueOf(groupIndex)));
					XUtils.putAttribute(detailTraces,
							XUtils.createAttribute("variant:count", Long.valueOf(groupCount)));
				}

				protected void done() {
					try {
						ProvidedObjectHelper.publish(context,
								groupName != null ? groupName : "Alignments Export as XLog with Extension", get(),
								XLog.class, true);
					} catch (InterruptedException | ExecutionException e) {
						ProMUIHelper.showErrorMessage(masterView.getMasterComponent(), e.toString(),
								"Could not export traces");
					} finally {
						masterView.getMasterList().afterUpdate();
					}
				}

			}.execute();

		}
	}
}