/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.cache;

import java.util.Arrays;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;

import nl.tue.astar.AStarThread;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;


public abstract class AbstractDataViewCache {

	protected final boolean activateDataViewCache;
	private final DataViewBuilder dataViewBuilder;
	private short[] currentExactView;
	private short[] tempDataView;

	public AbstractDataViewCache(DataViewBuilder dataViewBuilder, boolean activateDataViewCache, AbstractPDelegate<? extends Tail> delegate) {
		this.dataViewBuilder = dataViewBuilder;
		this.activateDataViewCache = activateDataViewCache;
		this.tempDataView = new short[1024];
	}
	
	public void setCurrentResult(Record alignmentResult) {	
		tempDataView = ensureSize(tempDataView, (alignmentResult.getBacktraceSize()+1) * 2); // max 2 shorts per step
		short i = 0;
		while (alignmentResult.getPredecessor() != null) {
			Record rec = alignmentResult;
			if (rec.getModelMove() != AStarThread.NOMOVE) {
				// Model Moved
				short writeIndex = dataViewBuilder.buildWriteDataView(rec.getModelMove());
				if (writeIndex != -1) {
					//TODO what to do with invisible model moves that write via virtual variables?
					// probably cache and graph search does not work then ...
					if (rec.getMovedEvent() != AStarThread.NOMOVE) {
						// Move in both: add index of transition together with mapped event  
						tempDataView[i++] = dataViewBuilder.buildEventDataView(writeIndex, (short)rec.getMovedEvent());
					} else {
						// Move in model: add index of transition with write operation
						tempDataView[i++] = writeIndex;
					}
				}

				short guardIndex = dataViewBuilder.buildGuardDataView(rec.getModelMove());
				if (guardIndex != -1) {
					// Move in model or move in both & guarded transition: add which guard was checked
					tempDataView[i++] = guardIndex;
				}
			}
			alignmentResult = alignmentResult.getPredecessor();
		}
		currentExactView = Arrays.copyOf(tempDataView, i);
	}

	private static short[] ensureSize(short[] a, int newLength) {
		if (a.length >= newLength) {
			return a;
		} else {
			return Arrays.copyOf(a, newLength+1);
		}
	}

	protected short[] getExactDataView() {
		return currentExactView;
	}
	
	public abstract int size();
	
}
