/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow;

/**
 * Thrown when the control flow alignment is unreliable (i.e. max number of
 * states exceeded or any other limit hit)
 * 
 * @author F. Mannhardt
 * 
 */
public class UnreliableControlFlowAlignmentException extends ControlFlowAlignmentException {

	public enum UnreliableReason {
		CANCELLED, MAXCOST, IMPOSSIBLE
	}

	private static final long serialVersionUID = -925650318028848655L;
	
	private final ControlFlowAlignmentResult unreliableResult;
	private final UnreliableReason reason;

	public UnreliableControlFlowAlignmentException(ControlFlowAlignmentResult unreliableResult, UnreliableReason reason) {
		super();
		this.unreliableResult = unreliableResult;
		this.reason = reason;
	}

	public UnreliableControlFlowAlignmentException(String message, ControlFlowAlignmentResult unreliableResult, UnreliableReason reason) {
		super(message);
		this.unreliableResult = unreliableResult;
		this.reason = reason;
	}

	public ControlFlowAlignmentResult getUnreliableResult() {
		return unreliableResult;
	}

	public UnreliableReason getReason() {
		return reason;
	}

	public String toString() {
		return super.toString() + ", Reason: "+ reason;
	}

}
