package org.processmining.plugins.balancedconformance.controlflow.override;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;

import gnu.trove.set.TShortSet;
import nl.tue.astar.AStarThread;

public abstract class AbstractDRecord implements DRecord {
	
	private static final DRecord[] EMPTY_ARRAY = new DRecord[0];

	public DRecord[] getHistory() {
		if (getBacktraceSize() < 0) {
			return EMPTY_ARRAY;
		}
		DRecord[] history = new DRecord[getBacktraceSize() + 1];
		DRecord r = this;
		int i = history.length;
		while (r.getPredecessor() != null) {
			history[--i] = r;
			r = r.getPredecessor();
		}
		return history;
	}

	public static void printRecord(AbstractPDelegate<?> delegate, int trace, DRecord r) {
		DRecord[] history = r.getHistory();

		for (int i = 0; i < history.length; i++) {
			r = history[i];
			String s = "(";
			short act = (short) delegate.getActivityOf(trace, r.getMovedEvent());

			if (r.getModelMove() == AStarThread.NOMOVE) {
				s += "_";
			} else {
				short m = (short) r.getModelMove();
				s += "(" + m + ")";
				// t is either a transition in the model, or AStarThread.NOMOVE
				TShortSet acts = delegate.getActivitiesFor(m);
				if (acts == null || acts.isEmpty() || !acts.contains(act)) {
					s += delegate.getTransition(m);
				} else {
					s += delegate.getEventClass(act);
				}
			}

			s += ",";
			// r.getLogEvent() is the event that was moved, or AStarThread.NOMOVE
			if (r.getMovedEvent() == AStarThread.NOMOVE) {
				s += "_";
			} else {
				assert (act >= 0 || act < 0);
				s += "(" + r.getMovedEvent() + ")" + delegate.getEventClass(act);
			}
			s += ") " + r.toString();
			s += (i < history.length - 1 ? " --> " : " cost: " + (r.getCostSoFar()));
			System.out.print(s);
		}
		System.out.println();
	}
}
