package org.processmining.plugins.balancedconformance.controlflow.override.queues.noduplicate;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;

import com.google.common.primitives.Ints;

import it.unimi.dsi.fastutil.Hash.Strategy;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Tail;

public final class ClosestSyncMovesDataPriorityQueue extends DataFastLookupPriorityQueue {

	private ClosestSyncMovesDataPriorityQueue(int initialCapacity, Strategy<DRecord> strategy,
			AbstractPDelegate<? extends Tail> delegate, int maxCost) {
		super(initialCapacity, strategy, delegate, maxCost);
	}

	private ClosestSyncMovesDataPriorityQueue(int initialCapacity, Strategy<DRecord> strategy,
			AbstractPDelegate<? extends Tail> delegate) {
		super(initialCapacity, strategy, delegate);
	}

	protected boolean isBetter(DRecord r1, DRecord r2) {
		// retrieve stored cost
		int c1 = r1.getTotalIntegerCost();
		int c2 = r2.getTotalIntegerCost();
		
		if (Ints.compare(c1, c2) < 0 
				|| (Ints.compare(c1, c2) == 0 && hasCloserSyncMoves(r1,r2))) {
			return true;
		}
		return false;
	}

	private boolean hasCloserSyncMoves(DRecord r1, DRecord r2) {
		int distanceBetweenSyncMoves1 = getSyncMoveDistance(r1);
		int distanceBetweenSyncMoves2 = getSyncMoveDistance(r2);
		return Ints.compare(distanceBetweenSyncMoves1, distanceBetweenSyncMoves2) < 0;
	}

	private int getSyncMoveDistance(DRecord record) {
		int totalDistance = 0;
		int index = record.getBacktraceSize();
		int lastSyncIndex = -1;
		while (record.getPredecessor() != null) {			
			if (record.getModelMove() != AStarThread.NOMOVE && record.getMovedEvent() != AStarThread.NOMOVE) {				
				if (lastSyncIndex != -1) {
					totalDistance += lastSyncIndex-index;
				}
				lastSyncIndex = index;	
			}
			record = record.getPredecessor();
			index--;
		}
		// Found synchronous move
		return totalDistance;
	}

}
