package org.processmining.plugins.balancedconformance.controlflow.override.storage;

import gnu.trove.map.TObjectIntMap;

import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import nl.tue.astar.Delegate;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.impl.JavaCollectionStorageHandler;
import nl.tue.astar.impl.State;

import org.processmining.plugins.astar.petrinet.impl.PHead;

public final class JavaCollectionStorageHandlerRWLock<T extends Tail> extends JavaCollectionStorageHandler<PHead, T> {

	private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock();
	private final Lock readLock = lock.readLock();
	private final Lock writeLock = lock.writeLock();

	public JavaCollectionStorageHandlerRWLock(Delegate<PHead, T> delegate, TObjectIntMap<PHead> head2int,
			List<State<PHead, T>> stateList) {
		super(delegate, head2int, stateList);
	}

	@Override
	public void storeStateForRecord(State<PHead, T> state, Record newRec) {
		int newIndex;
		synchronized (stateList) {
			newIndex = stateList.size() + 1;
			stateList.add(state);		
		}
		writeLock.lock();
		try {
			head2int.put(state.getHead(), newIndex);
		} finally {
			writeLock.unlock();
		}
		newRec.setState(newIndex - 1);
	}

	@Override
	public long getIndexOf(PHead head) {
		readLock.lock();
		try {
			return head2int.get(head) - 1;
		} finally {
			readLock.unlock();
		}
	}

}