package org.processmining.plugins.balancedconformance.dataflow.datastate;

public abstract class AbstractDataStateImpl implements DataState {

	protected abstract void putInternal(String varName, long value);

	protected abstract Long tryGetInternal(String varName);

	protected abstract Long tryGetPrimeOrNormalInternal(String varName);

	private final long getInternal(String varName) {
		Long result = tryGetInternal(varName);
		if (result == null) {
			throw new RuntimeException("DataState does not contain variable " + varName);
		} else {
			return result;
		}
	}

	public final void putDouble(String varName, double value) {
		putInternal(varName, convertFromDouble(value));
	}

	public final void putLong(String varName, long value) {
		putInternal(varName, value);
	}

	public final long getLong(String varName) {
		return getInternal(varName);
	}

	public final double getDouble(String varName) {
		return Double.longBitsToDouble(getInternal(varName));
	}

	public final Double tryGetDouble(String varName) {
		Long result = tryGetInternal(varName);
		if (result != null) {
			return convertToDouble(result);	
		} else {
			return null;
		}
	}

	public final Long tryGetLong(String varName) {
		return tryGetInternal(varName);
	}

	public final Double tryGetPrimeOrNormalDouble(String varName) {
		Long result = tryGetPrimeOrNormalInternal(varName);
		if (result != null) {
			return convertToDouble(result);
		} else {
			return null;
		}
	}

	public final Long tryGetPrimeOrNormalLong(String varName) {
		return tryGetPrimeOrNormalInternal(varName);
	}

	private static double convertToDouble(long l) {
		return Double.longBitsToDouble(l);
	}

	private static long convertFromDouble(double d) {
		return Double.doubleToLongBits(d);
	}

	protected final static boolean isPrimeVar(String varName) {
		return varName.charAt(varName.length() - 1) == '\'';
	}

}