/*
* Adapted from SolverFactoryGurobi.java of Java ILP
* 
* Java ILP is free software: you can redistribute it and/or modify it under
* the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
*
* Java ILP is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
* License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with Java ILP. If not, see http://www.gnu.org/licenses/.
*/ 
package org.processmining.plugins.balancedconformance.dataflow.lp.gurobi;

import org.processmining.plugins.balancedconformance.dataflow.MILPProblemBuilder;

import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import net.sf.javailp.AbstractSolverFactory;
import net.sf.javailp.Solver;

/**
 * Gurobi Solver Factory that re-uses the same GRBEnv for all Solvers as recommended for Gurobi
 * 
 * @author F. Mannhardt
 *
 */
public class DataAlignmentSolverFactoryGurobi extends AbstractSolverFactory implements AutoCloseable  {

	private GRBEnv env;

	public DataAlignmentSolverFactoryGurobi() {
		super();
		try {
			env = new GRBEnv();
			env.set(GRB.DoubleParam.IntFeasTol, MILPProblemBuilder.SOLVING_PRECISION);
			env.set(GRB.DoubleParam.OptimalityTol, MILPProblemBuilder.SOLVING_PRECISION);
			env.set(GRB.IntParam.OutputFlag, 0);
			//env.set(GRB.DoubleParam.TimeLimit, 360);
		} catch (GRBException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public Solver get() {
		return getInternal();
	}

	@Override
	protected Solver getInternal() {
		return new DataAlignmentGurobiSolver(env);
	}
	
	@Override
	public void close() {
		try {
			env.dispose();
		} catch (GRBException e) {
			throw new RuntimeException(e);
		}
	}

}
