/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.result;

import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.DataConformance.DataAlignment.DataAlignmentState;
import org.processmining.plugins.DataConformance.framework.ExecutionTrace;

public class BalancedDataAlignmentState extends DataAlignmentState {

	private final int maxCost;
	private final int dataCost;

	private final XTrace originalTrace;

	public BalancedDataAlignmentState(ExecutionTrace logTrace, ExecutionTrace processTrace, int overallCost,
			int dataCost, int maxCost, XTrace trace) {
		super(logTrace, processTrace, overallCost, false);
		this.dataCost = dataCost;
		this.maxCost = maxCost;
		this.originalTrace = trace;
	}

	public float getOverallFitness() {
		float fitness;
		if (getCost() == Float.MAX_VALUE || maxCost == 0.0) {
			// maxCost == 0.0 means that model and log share no common activity
			fitness = 0.0f;
		} else {
			fitness = 1 - (getCost() / maxCost);
		}
		// Negative fitness values do not make sense
		// this is a really bad alignment as its cost exceeds the maximum cost (all move on log, move on model with all data violations) 
		// on the shortest path through the model.
		return Math.max(0, fitness);
	}

	public int getDataCost() {
		return dataCost;
	}

	@Override
	public ExecutionTrace getProcessTracePrefix() {
		// avoid wrapping in unmodifiable collection
		return processTracePrefix;
	}

	@Override 
	public ExecutionTrace getLogTracePrefix() {
		// avoid wrapping in unmodifiable collection
		return logTracePrefix;
	}

	public XTrace getOriginalTrace() {
		return originalTrace;
	}

	public int getMaxCost() {
		return maxCost;
	}

}