package org.processmining.plugins.balancedconformance.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.fluxicon.slickerbox.factory.SlickerFactory;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;

public class ControlFlowCostUI extends JComponent {

	private static final long serialVersionUID = -1764028439791226196L;

	// default value 
	private static final int DEFCOSTMOVEONLOG = 1;
	private static final int DEFCOSTMOVEONMODEL = 1;

	// parameter-related GUI
	private Map<Transition, Integer> mapTrans2RowIndex = new HashMap<>();
	private DefaultTableModel transTableModel = null;
	private ProMTable promTable;

	private Map<XEventClass, Integer> mapXEvClass2RowIndex = new HashMap<>();
	private DefaultTableModel evClassTableModel = null;
	private ProMTable promEvClassTable;

	private boolean isUnmappedCostZero;
	
	private List<Transition> transCol;
	private TransEvClassMapping mapping;

	public ControlFlowCostUI(List<Transition> transCol, List<XEventClass> evClassCol, TransEvClassMapping mapping) {
		this(transCol, evClassCol, mapping, DEFCOSTMOVEONLOG, DEFCOSTMOVEONMODEL);
	}

	public ControlFlowCostUI(List<Transition> transCol, List<XEventClass> evClassCol, TransEvClassMapping mapping,
			int defaultCostMoveOnLog, int defaultCostMoveOnModel) {
		this.transCol = transCol;
		this.mapping = mapping;
		double size[][] = { { TableLayoutConstants.FILL },
				{ 80, 40, TableLayoutConstants.FILL, 35, TableLayoutConstants.FILL, 30, TableLayoutConstants.FILL, 30 } };
		TableLayout tl = new TableLayout(size);
		setLayout(tl);

		// label
		SlickerFactory slickerFactoryInstance = SlickerFactory.instance();
		add(slickerFactoryInstance
				.createLabel("<html><h1>Configure Cost for Control-Flow Violations</h1><p>Double click costs on table to change their values. Use only non-negative integers.</p></html>"),
				"0, 0, l, t");

		populateMoveOnModelPanel(transCol, defaultCostMoveOnModel);
		populateSetAllButton(String.valueOf(defaultCostMoveOnModel), transTableModel, "0, 3, c, t",
				"Set all costs (non-invisible only) above to ");

		populateMoveOnLogPanel(evClassCol, defaultCostMoveOnLog);
		populateSetAllButton(String.valueOf(defaultCostMoveOnLog), evClassTableModel, "0, 5, c, t",
				"Set all costs above to ");

	}

	private void populateMoveOnModelPanel(Collection<Transition> transitions, int defaultCostMoveOnModel) {
		// create table to map move on model cost
		Object[][] tableContent = new Object[transitions.size()][2];
		int rowCounter = 0;
		for (Transition trans : transitions) {
			tableContent[rowCounter] = new Object[] { trans.getLabel(),
					trans.isInvisible() ? 0 : defaultCostMoveOnModel };
			mapTrans2RowIndex.put(trans, rowCounter);
			rowCounter++;
		}
		transTableModel = new DefaultTableModel(tableContent, new Object[] { "Transition", "Move on Model Cost" }) {
			private static final long serialVersionUID = -6019224467802441949L;

			@Override
			public boolean isCellEditable(int row, int column) {
				return (column != 0);
			}

			@Override
			public void setValueAt(Object aValue, int row, int column) {
				try {
					Integer.parseInt((String) aValue);
					super.setValueAt(aValue, row, column);	
				} catch (NumberFormatException e) {
					ProMUIHelper.showWarningMessage(ControlFlowCostUI.this, "Please specify the cost as integer only. Could not parse " +  e.getMessage(), "Please specify the cost as integer only");
				}
			};
			
		};
		promTable = new ProMTable(transTableModel);
		promTable.setPreferredSize(new Dimension(700, 200));
		promTable.setMinimumSize(new Dimension(700, 200));
		add(promTable, "0, 2, c, t");
	}

	private void populateSetAllButton(String defaultCost, final DefaultTableModel tableModel, String addLocation,
			String label) {
		SlickerFactory factory = SlickerFactory.instance();

		final ProMTextField textField = new ProMTextField(defaultCost);
		textField.setMaximumSize(new Dimension(70, 20));
		//		textField.setMinimumSize(new Dimension(70,25));
		textField.setPreferredSize(new Dimension(70, 20));
		JButton setButton = factory.createButton("Set");
		setButton.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				try {
					int cost = Integer.parseInt(textField.getText().trim());
					if (cost >= 0) {
						for (int i = 0; i < tableModel.getRowCount(); i++) {
							if (tableModel == transTableModel) {
								if (!transCol.get(i).isInvisible()) {
									tableModel.setValueAt(cost, i, 1);
								}
							} else {
								tableModel.setValueAt(cost, i, 1);
							}
						}
					}
				} catch (Exception exc) {
					// no action is performed
				}
			}
		});

		JPanel bgPanel = new JPanel();
		bgPanel.setBackground(new Color(150, 150, 150));
		bgPanel.add(factory.createLabel(label));
		bgPanel.add(textField);
		bgPanel.add(setButton);
		add(bgPanel, addLocation);
	}

	/**
	 * Generate event class move on log panel
	 * 
	 * @param eventClassesName
	 * @param defaultCostMoveOnLog
	 */
	private void populateMoveOnLogPanel(Collection<XEventClass> eventClassesName, int defaultCostMoveOnLog) {
		// move on log cost (determined by the selection of event class in mapping)
		evClassTableModel = new DefaultTableModel() {
			private static final long serialVersionUID = 5238526181467190856L;

			@Override
			public boolean isCellEditable(int row, int column) {
				return (column != 0);
			};
			
			@Override
			public void setValueAt(Object aValue, int row, int column) {
				try {
					Integer.parseInt((String) aValue);
					super.setValueAt(aValue, row, column);	
				} catch (NumberFormatException e) {
					ProMUIHelper.showWarningMessage(ControlFlowCostUI.this, "Please specify the cost as integer only. Could not parse " +  e.getMessage(), "Please specify the cost as integer only");
				}
			};
			
		};

		mapXEvClass2RowIndex.clear();

		// move on log cost
		Object[][] evClassTableContent = new Object[eventClassesName.size()][2];
		int evClassRowCounter = 0;
		for (XEventClass evClass : eventClassesName) {
			if (!mapping.containsValue(evClass) && isUnmappedCostZero) {
				evClassTableContent[evClassRowCounter] = new Object[] { evClass, 0 };
			} else {
				evClassTableContent[evClassRowCounter] = new Object[] { evClass, defaultCostMoveOnLog };
			}
			mapXEvClass2RowIndex.put(evClass, evClassRowCounter);
			evClassRowCounter++;
		}
		evClassTableModel.setDataVector(evClassTableContent, new Object[] { "Event Class", "Move on Log Cost" });

		promEvClassTable = new ProMTable(evClassTableModel);
		promEvClassTable.setPreferredSize(new Dimension(700, 200));
		promEvClassTable.setMinimumSize(new Dimension(700, 200));
		add(promEvClassTable, "0, 4, c, t");
	}

	/**
	 * Get map from event class to cost of move on log
	 * 
	 * @return
	 */
	public Map<XEventClass, Integer> getMapEvClassToCost() {
		Map<XEventClass, Integer> mapEvClass2Cost = new HashMap<>();
		for (XEventClass evClass : mapXEvClass2RowIndex.keySet()) {
			int index = mapXEvClass2RowIndex.get(evClass);
			if (evClassTableModel.getValueAt(index, 1) instanceof Integer) {
				mapEvClass2Cost.put(evClass, (Integer) evClassTableModel.getValueAt(index, 1));
			} else {
				try {
					mapEvClass2Cost.put(evClass,
							Integer.parseInt(evClassTableModel.getValueAt(index, 1).toString().trim()));
				} catch (Exception exc) {
					mapEvClass2Cost.put(evClass, DEFCOSTMOVEONLOG);
				}

			}
		}
		return mapEvClass2Cost;
	}

	/**
	 * get penalty when move on model is performed
	 * 
	 * @return
	 */
	public Map<Transition, Integer> getTransitionWeight() {
		Map<Transition, Integer> costs = new HashMap<>();
		for (Transition trans : mapTrans2RowIndex.keySet()) {
			int index = mapTrans2RowIndex.get(trans);
			if (transTableModel.getValueAt(index, 1) instanceof Integer) {
				costs.put(trans, (Integer) transTableModel.getValueAt(index, 1));
			} else { // instance of other
				try {
					costs.put(trans, Integer.parseInt(transTableModel.getValueAt(index, 1).toString().trim()));
				} catch (Exception exc) {
					costs.put(trans, DEFCOSTMOVEONMODEL);
				}

			}
		}
		return costs;
	}

	/**
	 * get cost of doing synchronous moves
	 * 
	 * @return
	 */
	public Map<Transition, Integer> getSyncCost() {
		Map<Transition, Integer> costs = new HashMap<>(1);
		for (Transition transition : transCol) {
			costs.put(transition, 0);
		}
		return costs;
	}
}
